{$R-,B-}
unit skMHLjam;

interface
uses
     skMHL, skCommon;

const
 jmbAbsentJHR           = $2000;
 jmbAbsentJDX           = $2001;
 jmbAbsentJDT           = $2002;
 jmbCantOpenJHR         = $2003;
 jmbCantOpenJDX         = $2004;
 jmbCantOpenJDT         = $2005;
 jmbStreamProblems      = $2006;
 jmbWrongHeaderLocation = $2007;
 jmbWrongDataLocation   = $2008;
 jmbUnknownSubField     = $2009;
 jmbCantCreateJHR       = $200A;
 jmbCantCreateJDX       = $200B;
 jmbCantCreateJDT       = $200C;
 jmbAbsentJLR           = $200D;
 jmbCantOpenJLR         = $200E;
 jmbCantCreateJLR       = $200F;

 jamSource              = $0000;
 jamDestination         = $0001;
 jamFrom                = $0002;
 jamTo                  = $0003;
 jamMSGID               = $0004;
 jamREPLY               = $0005;
 jamSubject             = $0006;
 jamPID                 = $0007;
 jamVia                 = $0008;
 jamAttaches            = $0009;
 jamFreqs               = $000B;
 jamUnknown             = $07D0;
 jamSEENBY              = $07D1;
 jamPATH                = $07D2;
 jamFLAGS               = $07D3;

 jamaPrivate            = $00000004;
 jamaCrash              = $00000100;
 jamaReceived           = $00000008;
 jamaSent               = $00000010;
 jamaAttach             = $00002000;
 jamaTransit            = $00000002;
 jamaOrphan             = $00040000;
 jamaKill               = $00000020;
 jamaLocal              = $00000001;
 jamaHold               = $00000080;
 jamaFRq                = $00001000;
 jamaRRq                = $00010000;
 jamaARq                = $00020000;

 jamaTypeLocal          = $00800000;
 jamaTypeEcho           = $01000000;
 jamaTypeNet            = $02000000;

 JamSubBufferSize       = $3000;

type
 TJamBaseHeader = record
  Signature: Array[1..4] Of Char;
  Created: LongInt;
  ModCounter: LongInt;
  ActiveMsgs: LongInt;
  PwdCRC: LongInt;
  BaseMsgNum: LongInt;
  Extra: Array[1..1000] Of Char;
 end;

 TJamMessageHeaderFirst = record
  Signature: array[1..4] of char;
  Rev: Word;
  Resvd: Word;
  SubFieldLen: LongInt;
  TimesRead: LongInt;
  MsgIdCrc: LongInt;
  ReplyCrc: LongInt;
  ReplyTo: LongInt;
  ReplyFirst: LongInt;
  ReplyNext: LongInt;
  DateWritten: LongInt;
  DateRcvd: LongInt;
  DateArrived: LongInt;
  MsgNum: LongInt;
  Attr1: LongInt;
  Attr2: LongInt;
  TextOfs: LongInt;
  TextLen: LongInt;
  PwdCrc: LongInt;
  Cost: LongInt;
 end;

 TJamIndex = record
  MsgToCrc: LongInt;
  HdrLoc: LongInt;
 end;

 TJamLastRead = record
  NameCrc: LongInt;
  UserNum: LongInt;
  LastRead: LongInt;
  HighRead: LongInt;
 end;

 TJamSubBuffer = array[1..JamSubBufferSize] of char;

 PSubField = ^TSubField;
 TSubField = record
  LoID, HiID: Word;
  Length: Longint;
  Data: array[1..1000] of char;
 end;

 TJamMessageHeader = record
  JamHeader: TJamMessageHeaderFirst;
  SubBuffer: TJamSubBuffer;
 end;

 TJamInfo = record
  SourceAddress: TAddress;
  DestinationAddress: TAddress;
  SourceName: String;
  DestinationName: String;
  Subject: String;
 end;

 PJamMessageBase = ^TJamMessageBase;
 TJamMessageBase = object(TMessageBase)
  constructor Init;
  destructor Done; virtual;
  function Open(const Path: String): Boolean; virtual;
  function Create(const Path: String): Boolean; virtual;
  procedure Close; virtual;
  function Exists(Message: Longint): Boolean; virtual;
  function Current: Longint; virtual;
  function GetLocation: Longint; virtual;
  procedure SetLocation(Location: Longint); virtual;
  function OpenMessage: Boolean; virtual;
  function CloseMessage: Boolean; virtual;
  function GetHighest: Longint; virtual;
  function GetCount: Longint; virtual;
  function GetFrom: String; virtual;
  function GetTo: String; virtual;
  function GetSubject: String; virtual;
  procedure SetFrom(const S: String); virtual;
  procedure SetTo(const S: String); virtual;
  procedure SetSubject(const S: String); virtual;
  function GetAttribute(Attribute: Longint): Boolean; virtual;
  procedure SetAttribute(Attribute: Longint; Enable: Boolean); virtual;
  procedure GetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure GetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  function WriteMessage: Boolean; virtual;
  function CreateNewMessage: Boolean; virtual;
  function KillMessage: Boolean; virtual;
 private
  BasePath: String;
  HeaderLink: PMessageBaseStream;
  IndexLink: PMessageBaseStream;
  DataLink: PMessageBaseStream;
  LRLink: PMessageBaseStream;
  JamDataPosition: Longint;
  JamDataSize: Longint;
  JamBaseHeader: TJamBaseHeader;
  JamMessageHeader: TJamMessageHeader;
  JamMessageHeaderPosition: Longint;
  JamMessageHeaderSize: Longint;
  JamIndex: TJamIndex;
  JamIndexPosition: Longint;
  JamInfo: TJamInfo;
  function GetSubField(const Number: Longint; var ID: Longint; var S: String): Boolean;
  procedure AddSubField(const ID: Longint; const S: String);
  function MapAttribute(var Attribute: Longint): Boolean;
  function StrCrc32(const S: String): Longint;
  procedure ResetAll;
 end;

implementation

constructor TJamMessageBase.Init;
 begin
  inherited Init;
  BasePath:='';
  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;
  CurrentMessage:=0;
  MessageText:=Nil;
 end;

destructor TJamMessageBase.Done;
 begin
  CloseMessage;
  Close;
  inherited Done;
 end;

function TJamMessageBase.Open(const Path: String): Boolean;
 var
  Index: TJamIndex;
  TempStream: PMessageBaseStream;
  TempSize: Longint;
 begin
  if HeaderLink <> Nil then
   Close;

  Open:=False;
  BasePath:=Path;

  if BasePath[Length(BasePath)] <> '.' then
   BasePath:=Concat(BasePath, '.');

  if not FileExists(BasePath + 'JHR') then
   begin
    Status:=jmbAbsentJHR;

    Exit;
   end;

  if not FileExists(BasePath + 'JDX') then
   begin
    Status:=jmbAbsentJDX;

    Exit;
   end;

  if not FileExists(BasePath + 'JDT') then
   begin
    Status:=jmbAbsentJDT;

    Exit;
   end;

  if not FileExists(BasePath + 'JLR') then
   begin
    Status:=jmbAbsentJLR;

    Exit;
   end;

  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;

  repeat
   HeaderLink:=CreateMessageBaseFileStream(BasePath + 'JHR', smOpen);

   if HeaderLink^.Status <> smOk then
    begin
     Status:=jmbCantOpenJHR;

     Break;
    end;

   IndexLink:=CreateMessageBaseFileStream(BasePath + 'JDX', smOpen);

   if IndexLink^.Status <> smOk then
    begin
     Status:=jmbCantOpenJDX;

     Break;
    end;

   DataLink:=CreateMessageBaseFileStream(BasePath + 'JDT', smOpen);

   if DataLink^.Status <> smOk then
    begin
     Status:=jmbCantOpenJDT;

     Break;
    end;

   LRLink:=CreateMessageBaseFileStream(BasePath + 'JLR', smOpen);

   if LRLink^.Status <> smOk then
    begin
     Status:=jmbCantOpenJLR;

     Break;
    end;

   HeaderLink^.Seek(0);

   HeaderLink^.Read(JamBaseHeader, SizeOf(JamBaseHeader));

   if IndexLink^.GetSize div 8 <> JamBaseHeader.ActiveMsgs then
    begin
     IndexLink^.Seek(0);

     TempStream:=CreateMessageBaseMemoryStream(IndexLink^.GetSize);
     TempSize:=IndexLink^.GetSize;

     while TempSize <> 0 do
      begin
       IndexLink^.Read(Index, SizeOf(Index));

       if Index.HdrLoc <> -1 then
        TempStream^.Write(Index, SizeOf(Index));

       Dec(TempSize, SizeOf(Index));
      end;

     IndexLink^.Seek(0);
     TempStream^.Seek(0);

     IndexLink^.CopyFrom(TempStream^, TempStream^.GetSize);
     IndexLink^.Truncate;
     IndexLink^.Flush;

     Dispose(TempStream, Done);
    end;

   Open:=True;
   Opened:=True;
   Exit;
  until True;

  if HeaderLink <> Nil then
   Dispose(HeaderLink, Done);

  if IndexLink <> Nil then
   Dispose(IndexLink, Done);

  if DataLink <> Nil then
   Dispose(DataLink, Done);

  if LRLink <> Nil then
   Dispose(LRLink, Done);

  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;

  BasePath:='';
 end;

function TJamMessageBase.Create(const Path: String): Boolean;
 var
  DateTime: TMessageBaseDateTime;
 begin
  Create:=False;
  BasePath:=Path;

  if BasePath[Length(BasePath)] <> '.' then
   BasePath:=Concat(BasePath, '.');

  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;

  repeat
   HeaderLink:=CreateMessageBaseFileStream(BasePath + 'JHR', smCreate);

   if HeaderLink^.Status <> smOk then
    begin
     Status:=jmbCantCreateJHR;
     Break;
    end;

   IndexLink:=CreateMessageBaseFileStream(BasePath + 'JDX', smCreate);

   if IndexLink^.Status <> smOk then
    begin
     Status:=jmbCantCreateJDX;
     Break;
    end;

   DataLink:=CreateMessageBaseFileStream(BasePath + 'JDT', smCreate);

   if DataLink^.Status <> smOk then
    begin
     Status:=jmbCantCreateJDT;
     Break;
    end;

   LRLink:=CreateMessageBaseFileStream(BasePath + 'JLR', smCreate);

   if LRLink^.Status <> smOk then
    begin
     Status:=jmbCantCreateJLR;
     Break;
    end;

   ResetAll;

   FillChar(JamBaseHeader, SizeOf(JamBaseHeader), 0);

   JamBaseHeader.Signature[1]:='J';
   JamBaseHeader.Signature[2]:='A';
   JamBaseHeader.Signature[3]:='M';
   JamBaseHeader.Signature[4]:=#0;

   GetCurrentMessageBaseDateTime(DateTime);
   MessageBaseDateTimeToUnixDateTime(DateTime, JamBaseHeader.Created);

   JamBaseHeader.ModCounter:=0;
   JamBaseHeader.ActiveMsgs:=0;
   JamBaseHeader.PwdCRC:=-1;
   JamBaseHeader.BaseMsgNum:=1;

   HeaderLink^.Seek(0);
   HeaderLink^.Write(JamBaseHeader, SizeOf(JamBaseHeader));

   ResetAll;

   Create:=True;
   Opened:=True;

   Exit;
  until True;

  if HeaderLink <> Nil then
   Dispose(HeaderLink, Done);

  if IndexLink <> Nil then
   Dispose(IndexLink, Done);

  if DataLink <> Nil then
   Dispose(DataLink, Done);

  if LRLink <> Nil then
   Dispose(LRLink, Done);

  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;
 end;

procedure TJamMessageBase.Close;
 begin
  if (HeaderLink = Nil) or (IndexLink = Nil) or (DataLink = Nil) or (LRLink = Nil) then Exit;
  ResetAll;

  HeaderLink^.Seek(0);
  HeaderLink^.Write(JamBaseHeader, SizeOf(JamBaseHeader));
  HeaderLink^.Flush;

  IndexLink^.Flush;
  DataLink^.Flush;
  LRLink^.Flush;

  Dispose(DataLink, Done);
  Dispose(IndexLink, Done);
  Dispose(HeaderLink, Done);
  Dispose(LRLink, Done);

  HeaderLink:=Nil;
  IndexLink:=Nil;
  DataLink:=Nil;
  LRLink:=Nil;

  Opened:=False;
 end;

function TJamMessageBase.Exists(Message: Longint): Boolean;
 begin
  Exists:=(Message >= JamBaseHeader.BaseMsgNum) and (Message < JamBaseHeader.BaseMsgNum + JamBaseHeader.ActiveMsgs);
 end;

function TJamMessageBase.Current: Longint;
 begin
  Current:=CurrentMessage;
 end;

function TJamMessageBase.GetLocation: Longint;
 begin
  GetLocation:=CurrentMessage;
 end;

procedure TJamMessageBase.SetLocation(Location: Longint);
 begin
  CurrentMessage:=Location;
 end;

function TJamMessageBase.OpenMessage: Boolean;
 var
  Size, Count, SubFieldType: Longint;
  S: String;
 begin
  if MessageText <> Nil then
   CloseMessage;

  ResetAll;

  OpenMessage:=False;

  JamIndexPosition:=(CurrentMessage - JamBaseHeader.BaseMsgNum) * SizeOf(JamIndex);

  IndexLink^.Seek(JamIndexPosition);

  if IndexLink^.Status <> smOk then
   begin
    Status:=jmbStreamProblems;

    Exit;
   end;

  IndexLink^.Read(JamIndex, SizeOf(JamIndex));

  if (JamIndex.HdrLoc < 0) or (JamIndex.HdrLoc > HeaderLink^.GetSize) then
   begin
    Status:=jmbWrongHeaderLocation;

    Exit;
   end;

  HeaderLink^.Seek(JamIndex.HdrLoc);

  if HeaderLink^.Status <> smOk then
   begin
    Status:=jmbStreamProblems;

    Exit;
   end;

  Size:=HeaderLink^.GetSize - HeaderLink^.GetPos;

  if Size > SizeOf(JamMessageHeader) then
   Size:=SizeOf(JamMessageHeader);

  HeaderLink^.Read(JamMessageHeader, Size);

  if HeaderLink^.Status <> smOk then
   begin
    Status:=jmbStreamProblems;

    Exit;
   end;

  if (JamMessageHeader.JamHeader.TextOfs < 0) or (JamMessageHeader.JamHeader.TextOfs > DataLink^.GetSize) then
   begin
    Status:=jmbWrongDataLocation;

    Exit;
   end;

  FillChar(JamInfo, SizeOf(JamInfo), 0);

  Count:=1;

  while GetSubField(Count, SubFieldType, S) do
   begin
    case SubFieldType of
     jamSource: StrToAddress(S, JamInfo.SourceAddress);
     jamDestination: StrToAddress(S, JamInfo.DestinationAddress);
     jamFrom: JamInfo.SourceName:=S;
     jamTo: JamInfo.DestinationName:=S;
     jamSubject: JamInfo.Subject:=S;
     jamAttaches: if GetAttribute(maAttach) then JamInfo.Subject:=S;
     jamFreqs: if GetAttribute(maFRq) then JamInfo.Subject:=S;
    end;

    Inc(Count);
   end;

  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);
  DataLink^.Seek(JamMessageHeader.JamHeader.TextOfs);
  MessageText^.Seek(0);

  Count:=1;
  while GetSubField(Count, SubFieldType, S) do
   begin
    case SubFieldType of
     jamMSGID: PutString('MSGID: ' + S);
     jamREPLY: PutString('REPLY: ' + S);
     jamPID: PutString('PID: ' + S);
     jamUnknown: PutString('' + S);
     jamFLAGS: PutString('FLAGS: ' + S);
    end;
    Inc(Count);
   end;

  MessageText^.CopyFrom(DataLink^, JamMessageHeader.JamHeader.TextLen);

  Count:=1;
  while GetSubField(Count, SubFieldType, S) do
   begin
    if SubFieldType = jamVia then
     PutString('Via ' + S);

    Inc(Count);
   end;

  Count:=1;
  while GetSubField(Count, SubFieldType, S) do
   begin
    if SubFieldType = jamSEENBY then
     PutString('SEEN-BY: ' + S);

    Inc(Count);
   end;

  Count:=1;
  while GetSubField(Count, SubFieldType, S) do
   begin
    if SubFieldType = jamPATH then
     PutString('PATH: ' + S);

    Inc(Count);
   end;

  if JamInfo.SourceAddress.Zone <> 0 then
   SetFromAddress(JamInfo.SourceAddress, False);

  if JamInfo.DestinationAddress.Zone <> 0 then
   SetToAddress(JamInfo.DestinationAddress);

  MessageText^.Seek(0);
  OpenMessage:=True;
 end;

function TJamMessageBase.CloseMessage: Boolean;
 begin
  CloseMessage:=False;

  if MessageText = Nil then
   Exit;

  CloseMessage:=True;

  Dispose(MessageText, Done);

  MessageText:=Nil;
 end;

function TJamMessageBase.GetHighest: Longint;
 begin
  GetHighest:=JamBaseHeader.BaseMsgNum + JamBaseHeader.ActiveMsgs;
 end;

function TJamMessageBase.GetCount: Longint;
 begin
  GetCount:=JamBaseHeader.ActiveMsgs;
 end;

function TJamMessageBase.GetFrom: String;
 begin
  GetFrom:=JamInfo.SourceName;
 end;

function TJamMessageBase.GetTo: String;
 begin
  GetTo:=JamInfo.DestinationName;
 end;

function TJamMessageBase.GetSubject: String;
 begin
  GetSubject:=JamInfo.Subject;
 end;

procedure TJamMessageBase.SetFrom(const S: String);
 begin
  JamInfo.SourceName:=S;
 end;

procedure TJamMessageBase.SetTo(const S: String);
 begin
  JamInfo.DestinationName:=S;
 end;

procedure TJamMessageBase.SetSubject(const S: String);
 begin
  JamInfo.Subject:=S;
 end;

function TJamMessageBase.GetAttribute(Attribute: Longint): Boolean;
 begin
  if not MapAttribute(Attribute) then
   begin
    GetAttribute:=False;

    Exit;
   end;

  GetAttribute:=JamMessageHeader.JamHeader.Attr1 and Attribute <> 0;
 end;

procedure TJamMessageBase.SetAttribute(Attribute: Longint; Enable: Boolean);
 begin
  if not MapAttribute(Attribute) then
   Exit;

  if Enable then
   JamMessageHeader.JamHeader.Attr1:=JamMessageHeader.JamHeader.Attr1 or Attribute
  else
   JamMessageHeader.JamHeader.Attr1:=JamMessageHeader.JamHeader.Attr1 and not Attribute
 end;

procedure TJamMessageBase.GetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 begin
  UnixDateTimeToMessageBaseDateTime(JamMessageHeader.JamHeader.DateWritten, DateTime);
 end;

procedure TJamMessageBase.GetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
  UnixDateTimeToMessageBaseDateTime(JamMessageHeader.JamHeader.DateArrived, DateTime);
 end;

procedure TJamMessageBase.SetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 begin
  MessageBaseDateTimeToUnixDateTime(DateTime, JamMessageHeader.JamHeader.DateWritten);
 end;

procedure TJamMessageBase.SetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
  MessageBaseDateTimeToUnixDateTime(DateTime, JamMessageHeader.JamHeader.DateArrived);
 end;

function TJamMessageBase.WriteMessage: Boolean;
 var
  Buffer: PMessageBaseStream;
  Line: PChar;
  B: System.Byte;
  S: String;
  LineLength: Byte absolute S;
 begin
  ResetAll;

  B:=13;

  {* Creating subfields and repacking message text *}
  JamMessageHeader.JamHeader.SubFieldLen:=0;
  GetMem(Line, MaxLineSize);
  Buffer:=CreateMessageBaseMemoryStream(MaxMessageSize);

  GetFromAndToAddress(JamInfo.SourceAddress, JamInfo.DestinationAddress);

  AddSubField(jamFrom, JamInfo.SourceName);
  AddSubField(jamSource, AddressToStr(JamInfo.SourceAddress));

  AddSubField(jamTo, JamInfo.DestinationName);

  if AddressCompare(JamInfo.DestinationAddress, NullAddress) <> 0 then
   AddSubField(jamDestination, AddressToStr(JamInfo.DestinationAddress))
  else
   AddSubField(jamDestination, AddressToStr(JamInfo.SourceAddress));

  if JamInfo.Subject <> '' then
   AddSubField(jamSubject, JamInfo.Subject);

  if GetAttribute(maAttach) then
   AddSubField(jamAttaches, JamInfo.Subject);

  if GetAttribute(maFRq) then
   AddSubField(jamFreqs, JamInfo.Subject);

  SetTextPos(0);
  while not EndOfMessage do
   begin
    GetStringPChar(Line, MaxLineSize);

    LineLength:=LenASCIIZ(Line) and $FF;

    Move(Line^, S[1], LineLength);
    if Copy(S, 1, 8) = 'MSGID: ' then
     begin
      AddSubField(jamMSGID, Copy(S, 9, 255));

      JamMessageHeader.JamHeader.MSGIDCrc:=StrCrc32(Copy(S, 9, 255));
     end else
    if Copy(S, 1, 8) = 'REPLY: ' then
     begin
      AddSubField(jamREPLY, Copy(S, 9, 255));

      JamMessageHeader.JamHeader.REPLYCrc:=StrCrc32(Copy(S, 9, 255));
     end else
    if Copy(S, 1, 6) = 'PID: ' then AddSubField(jamPID, Copy(S, 7, 255)) else
    if Copy(S, 1, 8) = 'FLAGS: ' then AddSubField(jamFLAGS, Copy(S, 9, 255)) else
    if Copy(S, 1, 5) = 'Via ' then AddSubField(jamVia, Copy(S, 6, 255)) else
    if Copy(S, 1, 9) = 'SEEN-BY: ' then AddSubField(jamSEENBY, Copy(S, 10, 255)) else
    if Copy(S, 1, 7) = 'PATH: ' then AddSubField(jamPATH, Copy(S, 8, 255)) else
    if Copy(S, 1, 5) = 'INTL' then else
    if Copy(S, 1, 5) = 'FMPT' then else
    if Copy(S, 1, 5) = 'TOPT' then else
    if Copy(S, 1, 1) = '' then AddSubField(jamUnknown, Copy(S, 2, 255)) else
     begin
      Buffer^.Write(Line^, LenASCIIZ(Line));
      Buffer^.Write(B, SizeOf(B));
     end;
   end;
  FreeMem(Line, MaxLineSize);

  Buffer^.Seek(0);

  {* Update message *}
  SetAttribute(jamaTypeEcho, GetFlag(afEchomail));
  SetAttribute(jamaTypeNet, GetFlag(afNetmail));
  SetAttribute(jamaTypeLocal, GetFlag(afLocal));

  {* Update index *}
  JamMessageHeaderPosition:=HeaderLink^.GetSize;
  JamIndex.MsgToCrc:=StrCrc32(JamInfo.DestinationName);
  JamIndex.HdrLoc:=JamMessageHeaderPosition;

  {* Write index *}
  IndexLink^.Seek(JamIndexPosition);
  IndexLink^.Write(JamIndex, SizeOf(JamIndex));

  {* Update header *}
  JamDataPosition:=DataLink^.GetSize;
  JamDataSize:=Buffer^.GetSize;

  JamMessageHeader.JamHeader.TextOfs:=JamDataPosition;
  JamMessageHeader.JamHeader.TextLen:=JamDataSize;
  Inc(JamBaseHeader.ModCounter);

  {* Write header *}
  HeaderLink^.Seek(JamMessageHeaderPosition);
  HeaderLink^.Write(JamMessageHeader, SizeOf(JamMessageHeader.JamHeader) + JamMessageHeader.JamHeader.SubFieldLen);

  {* Write text *}
  DataLink^.Seek(JamDataPosition);
  DataLink^.CopyFrom(Buffer^, JamDataSize);

  {* Flush'em *}
  IndexLink^.Flush;
  HeaderLink^.Flush;
  DataLink^.Flush;
  LRLink^.Flush;

  Dispose(Buffer, Done);

  WriteMessage:=True;
 end;

function TJamMessageBase.CreateNewMessage: Boolean;
 var
  DateTime: TMessageBaseDateTime;
 begin
  ResetAll;

  Inc(JamBaseHeader.ActiveMsgs);
  CurrentMessage:=JamBaseHeader.BaseMsgNum + GetCount;

  FillChar(JamMessageHeader, SizeOf(JamMessageHeader), 0);
  FillChar(JamInfo, SizeOf(JamInfo), 0);

  JamIndex.MsgToCrc:=-1;

  JamMessageHeader.JamHeader.Signature[1]:='J';
  JamMessageHeader.JamHeader.Signature[2]:='A';
  JamMessageHeader.JamHeader.Signature[3]:='M';
  JamMessageHeader.JamHeader.Signature[4]:=#0;

  JamMessageHeader.JamHeader.Rev:=1;

  JamIndexPosition:=IndexLink^.GetSize;

  JamMessageHeader.JamHeader.MSGIDCrc:=$FFFFFFFF;
  JamMessageHeader.JamHeader.REPLYCrc:=$FFFFFFFF;
  JamMessageHeader.JamHeader.PwdCrc:=$FFFFFFFF;

  JamMessageHeader.JamHeader.MsgNum:=CurrentMessage - 1;

  GetCurrentMessageBaseDateTime(DateTime);
  SetWrittenDateTime(DateTime);

  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);

  CreateNewMessage:=True;
 end;

function TJamMessageBase.KillMessage: Boolean;
 begin
  ResetAll;

  JamIndex.MsgToCrc:=-1;
  JamIndex.HdrLoc:=-1;

  IndexLink^.Seek(JamIndexPosition);
  IndexLink^.Write(JamIndex, SizeOf(JamIndex));

  Dec(JamBaseHeader.ActiveMsgs);
 end;

{* Private methods *}

function TJamMessageBase.GetSubField(const Number: Longint; var ID: Longint; var S: String): Boolean;
 var
  K, L: Longint;
  Length: Byte absolute S;
  SubField: PSubField;
 begin
  L:=0;
  K:=1;
  while (K <= JamMessageHeader.JamHeader.SubFieldLen) and (K < JamSubBufferSize) do
   begin
    SubField:=@JamMessageHeader.SubBuffer[K];

    Inc(K, SubField^.Length + 8);
    Inc(L);

    if L = Number then
     begin
      GetSubField:=True;
      ID:=SubField^.LoID;
      Length:=SubField^.Length and $FF;

      Move(SubField^.Data, S[1], Length);

      Exit;
     end;
   end;
  GetSubField:=False;
 end;

procedure TJamMessageBase.AddSubField(const ID: Longint; const S: String);
 var
  SubField: PSubField;
 begin
  SubField:=@JamMessageHeader.SubBuffer[JamMessageHeader.JamHeader.SubFieldLen + 1];
  SubField^.LoID:=ID;
  SubField^.HiID:=0;
  SubField^.Length:=Length(S);

  Move(S[1], SubField^.Data, SubField^.Length);

  Inc(JamMessageHeader.JamHeader.SubFieldLen, 8 + Length(S));
 end;

function TJamMessageBase.MapAttribute(var Attribute: Longint): Boolean;
 begin
  MapAttribute:=True;

  case Attribute of
   maPrivate: Attribute:=jamaPrivate;
   maCrash: Attribute:=jamaCrash;
   maReceived: Attribute:=jamaReceived;
   maSent: Attribute:=jamaSent;
   maAttach: Attribute:=jamaAttach;
   maTransit: Attribute:=jamaTransit;
   maOrphan: Attribute:=jamaOrphan;
   maKill: Attribute:=jamaKill;
   maLocal: Attribute:=jamaLocal;
   maHold: Attribute:=jamaHold;
   maFRq: Attribute:=jamaFRq;
   maRRq: Attribute:=jamaRRq;
   maARq: Attribute:=jamaARq;
  else
   MapAttribute:=False;
  end;
 end;

function TJamMessageBase.StrCrc32(const S: String): Longint;
 var
  I: Byte;
  L: Longint;
 begin
  L:=-1;

  for I:=1 to Length(S) do
   L:=UpdateCRC32(Byte(LoCase(S[I])), L);

  StrCrc32:=L;
 end;

procedure TJamMessageBase.ResetAll;
 begin
  IndexLink^.Reset;
  IndexLink^.Flush;

  DataLink^.Reset;
  DataLink^.Flush;

  HeaderLink^.Reset;
  HeaderLink^.Flush;

  LRLink^.Reset;
  LRLink^.Flush;
 end;

end.