{$B-}
unit skMHLmsg;

interface
uses
     skMHL, skCommon;

const
 fmbWrongPath           = $1000;
 fmbMessageNotFound     = $1001;
 fmbCannotCreateStream  = $1002;
 fmbAlreadyClosed       = $1003;
 fmbBufferOverflow      = $1004;

 MaxMapSize             = $F000;

type
 PHeader     = ^THeader;
 THeader     = Record
  FromUser   : Array[1..36] Of Char;
  ToUser     : Array[1..36] Of Char;
  Subject    : Array[1..72] Of Char;
  DateTime   : Array[1..20] Of Char;
  TimesRead  : System.Word;
  DestNode   : System.Word;
  OrigNode   : System.Word;
  Cost       : System.Word;
  OrigNet    : System.Word;
  DestNet    : System.Word;
  DateWritten: Longint;
  DateArrived: Longint;
  ReplyTo    : System.Word;
  Attr       : System.Word;
  ReplyNext  : System.Word;
 end;

 TFidoMap = array[0..MaxMapSize div 8 - 1] of Byte;

 PFidoMessageBase = ^TFidoMessageBase;
 TFidoMessageBase = object(TMessageBase)
  constructor Init;
  destructor Done; virtual;
  function Open(const Path: String): Boolean; virtual;
  function Create(const Path: String): Boolean; virtual;
  procedure Close; virtual;
  function Exists(Message: Longint): Boolean; virtual;
  function Current: Longint; virtual;
  function GetLocation: Longint; virtual;
  procedure SetLocation(Location: Longint); virtual;
  function OpenMessage: Boolean; virtual;
  function CloseMessage: Boolean; virtual;
  function GetHighest: Longint; virtual;
  function GetCount: Longint; virtual;
  function GetFrom: String; virtual;
  function GetTo: String; virtual;
  function GetSubject: String; virtual;
  procedure GetFromAddress(var Address: TAddress); virtual;
  procedure GetToAddress(var Address: TAddress); virtual;
  procedure GetFromAndToAddress(var FromAddress, ToAddress: TAddress); virtual;
  procedure SetFrom(const S: String); virtual;
  procedure SetTo(const S: String); virtual;
  procedure SetSubject(const S: String); virtual;
  procedure SetFromAddress(var Address: TAddress; const FreshMSGID: Boolean); virtual;
  procedure SetToAddress(var Address: TAddress); virtual;
  procedure SetFromAndToAddress(var FromAddress, ToAddress: TAddress); virtual;
  function GetAttribute(Attribute: Longint): Boolean; virtual;
  procedure SetAttribute(Attribute: Longint; Enable: Boolean); virtual;
  procedure GetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure GetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  function WriteMessage: Boolean; virtual;
  function CreateNewMessage: Boolean; virtual;
  function KillMessage: Boolean; virtual;
 private
  Directory: String;
  Map: TFidoMap;
  TotalMessages: Longint;
  HighMessage: Longint;
  Link: PMessageBaseStream;
  Header: THeader;
  procedure RebuildIndex;
  procedure SetMapDot(Num: Integer; Value: Boolean);
  function GetMapDot(Num: Integer): Boolean;
  function MapAttribute(var Attribute: Longint): Boolean;
 end;

implementation

const
 MapPos: Array[0..7] Of Byte = (1, 2, 4, 8, 16, 32, 64, 128);

 faPrivate           = $0001;
 faCrash             = $0002;
 faReceived          = $0004;
 faSent              = $0008;
 faAttach            = $0010;
 faTransit           = $0020;
 faOrphan            = $0040;
 faKill              = $0080;
 faLocal             = $0100;
 faHold              = $0200;
 faFRq               = $0800;
 faRRq               = $1000;
 faRRc               = $2000;
 faARq               = $4000;
 faURq               = $8000;

{* TFidoMessageBase *}

constructor TFidoMessageBase.Init;
 begin
  inherited Init;
  Link:=Nil;
 end;

destructor TFidoMessageBase.Done;
 begin
  CloseMessage;
  Close;
  inherited Done;
 end;

function TFidoMessageBase.Open(const Path: String): Boolean;
 begin
  Open:=False;
  Opened:=False;
  Directory:=Path;
  if Directory[Length(Directory)] <> '\' then Directory:=Concat(Directory, '\');
  if not DirectoryExists(Path) then
   begin
    Status:=fmbWrongPath;
    Exit;
   end;
  RebuildIndex;
  Open:=True;
  Opened:=True;
 end;

function TFidoMessageBase.Create(const Path: String): Boolean;
 begin
  Create:=False;
  Opened:=False;
  Directory:=Path;
  if Directory[Length(Directory)] <> '\' then Directory:=Concat(Directory, '\');
  if not CreateDirectory(Directory) then Exit;
  RebuildIndex;
  Create:=True;
  Opened:=True;
 end;

procedure TFidoMessageBase.Close;
 begin
  Directory:='';
  Opened:=False;
 end;

function TFidoMessageBase.Exists(Message: Longint): Boolean;
 begin
  Exists:=GetMapDot(Message);
 end;

function TFidoMessageBase.Current: Longint;
 begin
  Current:=CurrentMessage;
 end;

function TFidoMessageBase.GetLocation: Longint;
 begin
  GetLocation:=CurrentMessage;
 end;

procedure TFidoMessageBase.SetLocation(Location: Longint);
 begin
  CurrentMessage:=Location;
 end;

function TFidoMessageBase.OpenMessage: Boolean;
 begin
  OpenMessage:=False;
  if not Exists(CurrentMessage) then
   begin
    Status:=fmbMessageNotFound;
    Exit;
   end;
  if Link <> Nil then CloseMessage;
  Link:=CreateMessageBaseFileStream(Directory + LongToStr(CurrentMessage) + '.MSG', smOpen);
  if Link^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Link:=Nil;
    Status:=fmbCannotCreateStream;
    Exit;
   end;
  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);
  if MessageText^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Dispose(MessageText, Done);
    Link:=Nil;
    Status:=fmbCannotCreateStream;
    Exit;
   end;
  Link^.Seek(0);
  Link^.Read(Header, SizeOf(Header));
  MessageText^.CopyFrom(Link^, Link^.GetSize - Link^.GetPos - 1);
  MessageText^.Seek(0);
  OpenMessage:=True;
 end;

function TFidoMessageBase.CloseMessage: Boolean;
 begin
  CloseMessage:=False;
  if Link = Nil then
   begin
    Status:=fmbAlreadyClosed;
    Exit;
   end;
  Dispose(Link, Done);
  Dispose(MessageText, Done);
  Link:=Nil;
  CloseMessage:=True;
 end;

function TFidoMessageBase.GetHighest: Longint;
 begin
  GetHighest:=HighMessage;
 end;

function TFidoMessageBase.GetCount: Longint;
 begin
  GetCount:=TotalMessages;
 end;

function TFidoMessageBase.GetFrom: String;
 begin
  GetFrom:=FromASCIIZ(@Header.FromUser);
 end;

function TFidoMessageBase.GetTo: String;
 begin
  GetTo:=FromASCIIZ(@Header.ToUser);
 end;

function TFidoMessageBase.GetSubject: String;
 begin
  GetSubject:=FromASCIIZ(@Header.Subject);
 end;

procedure TFidoMessageBase.GetFromAddress(var Address: TAddress);
 begin
  Address.Zone:=0;
  Address.Net:=Header.OrigNet;
  Address.Node:=Header.OrigNode;
  Address.Point:=0;
  
  inherited GetFromAddress(Address);
 end;

{//mab}
procedure TFidoMessageBase.GetToAddress(var Address: TAddress);
 var
  TempAddress: TAddress;
 begin
  Address.Zone:=0;
  Address.Point:=0;

  inherited GetToAddress(Address);

  Address.Net:=Header.DestNet;
  Address.Node:=Header.DestNode;

  if Address.Zone = 0 then
   begin
    GetFromAddress(TempAddress);
    Address.Zone:=TempAddress.Zone;
   end;
 end;
{//end}

procedure TFidoMessageBase.GetFromAndToAddress(var FromAddress, ToAddress: TAddress);
 begin
  FromAddress.Zone:=0;
  FromAddress.Net:=Header.OrigNet;
  FromAddress.Node:=Header.OrigNode;
  FromAddress.Point:=0;
  ToAddress.Zone:=0;
  ToAddress.Net:=Header.DestNet;
  ToAddress.Node:=Header.DestNode;
  ToAddress.Point:=0;
  inherited GetFromAndToAddress(FromAddress, ToAddress);
 end;

procedure TFidoMessageBase.SetFrom(const S: String);
 begin
  ToASCIIZ(S, @Header.FromUser);
 end;

procedure TFidoMessageBase.SetTo(const S: String);
 begin
  ToASCIIZ(S, @Header.ToUser);
 end;

procedure TFidoMessageBase.SetSubject(const S: String);
 begin
  ToASCIIZ(S, @Header.Subject);
 end;

procedure TFidoMessageBase.SetFromAddress(var Address: TAddress; const FreshMSGID: Boolean);
 begin
  inherited SetFromAddress(Address, FreshMSGID);
  Header.OrigNet:=Address.Net;
  Header.OrigNode:=Address.Node;
 end;

procedure TFidoMessageBase.SetToAddress(var Address: TAddress);
 begin
  inherited SetToAddress(Address);
  Header.DestNet:=Address.Net;
  Header.DestNode:=Address.Node;
 end;

procedure TFidoMessageBase.SetFromAndToAddress(var FromAddress, ToAddress: TAddress);
 begin
  inherited SetFromAndToAddress(FromAddress, ToAddress);
  Header.OrigNet:=FromAddress.Net;
  Header.OrigNode:=FromAddress.Node;
  Header.DestNet:=ToAddress.Net;
  Header.DestNode:=ToAddress.Node;
 end;

function TFidoMessageBase.GetAttribute(Attribute: Longint): Boolean;
 begin
  if not MapAttribute(Attribute) then
   begin
    GetAttribute:=False;
    Exit;
   end;
  GetAttribute:=Header.Attr and Attribute = Attribute;
 end;

procedure TFidoMessageBase.SetAttribute(Attribute: Longint; Enable: Boolean);
 begin
  if not MapAttribute(Attribute) then Exit;
  if Enable then
   Header.Attr:=Header.Attr or Attribute
  else
   Header.Attr:=Header.Attr and not Attribute;
 end;

{ Warning!
  TFidoMessageBase.GetWrittenDateTime was tested only with format of
  keeping of date named 'Fido'. Work with other formats of keeping of
  date is not guaranteed.
}
procedure TFidoMessageBase.GetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 var
  S: String;
  K: Byte;
 begin
  FillChar(DateTime, SizeOf(DateTime), 0);
  S:=FromASCIIZ(@Header.DateTime);
  if S[3] = ' ' then
   if S[11] = ' ' then {Fido - "DD MMM YY  HH:MM:SS"}
    begin
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 11);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     StrToWord(ExtractWord(3, S, [':', ' ']), DateTime.Sec);
    end
   else {Opus - "DD MMM YY HH:MM:SS"}
    begin
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 10);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     StrToWord(ExtractWord(3, S, [':', ' ']), DateTime.Sec);
    end
  else
   if S[4] = ' ' then {SeaDog - "DOW DD MMM YY HH:MM"}
    begin
     Delete(S, 1, 4);
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 10);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     DateTime.Sec:=0;
    end
   else
    if S[3] = '-' then {Wierd - "DD-MM-YYYY HH:MM:SS"}
     begin
      for K:=1 to Length(S) do
       if S[K] in ['-', ':'] then S[K]:=' ';
      StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
      StrToWord(ExtractWord(2, S, [' ']), DateTime.Month);
      StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
      StrToWord(ExtractWord(4, S, [' ']), DateTime.Hour);
      StrToWord(ExtractWord(5, S, [' ']), DateTime.Min);
      StrToWord(ExtractWord(6, S, [' ']), DateTime.Sec);
     end;
  if DateTime.Year < 1900 then
   Inc(DateTime.Year, 1900);
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  ???DateTimeToMessageBaseDateTime(Header.DateWritten, DateTime);}
 end;

procedure TFidoMessageBase.GetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  ???DateTimeToMessageBaseDateTime(Header.DateArrived, DateTime);}
 end;

procedure TFidoMessageBase.SetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 function LeadingZero(Number: Word): String;
  var
   S: String[2];
  begin
   Str(Number, S);
   if S[0] = #1 then
    S:=Concat('0', S);
   LeadingZero:=S;
  end;
 begin
  ToASCIIZ(LeadingZero(DateTime.Day) + ' ' +
           MonthNumberToMonthString(DateTime.Month) + ' ' +
           LeadingZero(DateTime.Year - 1900) + '  ' +
           LeadingZero(DateTime.Hour) + ':' +
           LeadingZero(DateTime.Min) + ':' +
           LeadingZero(DateTime.Sec), @Header.DateTime);
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  MessageBaseDateTimeTo???DateTime(DateTime, Header.DateWritten);}
 end;

procedure TFidoMessageBase.SetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  MessageBaseDateTimeTo???DateTime(DateTime, Header.DateArrived);}
 end;

function TFidoMessageBase.WriteMessage: Boolean;
 var
  B: System.Byte;
 begin
  WriteMessage:=True;
  Link^.Reset;
  Link^.Seek(0);
  Link^.Write(Header, SizeOf(Header));
  MessageText^.Seek(0);
  Link^.CopyFrom(MessageText^, MessageText^.GetSize);
  B:=0;
  Link^.Write(B, SizeOf(B));
  Link^.Truncate;
  Link^.Flush;
 end;

function TFidoMessageBase.CreateNewMessage: Boolean;
 begin
  CreateNewMessage:=False;
  CurrentMessage:=GetHighest + 1;
  Inc(HighMessage);
  SetMapDot(CurrentMessage, True);
  if Link <> Nil then CloseMessage;
  Link:=CreateMessageBaseFileStream(Directory + LongToStr(CurrentMessage) + '.MSG', smCreate);
  if Link^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Link:=Nil;
    Exit;
   end;
  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);
  if MessageText^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Dispose(MessageText, Done);
    Link:=Nil;
    Exit;
   end;
  FillChar(Header, SizeOf(Header), 0);
  CreateNewMessage:=True;
 end;

function TFidoMessageBase.KillMessage: Boolean;
 begin
  KillMessage:=False;
  if not Exists(CurrentMessage) then Exit;
  if not KillFile(Directory + LongToStr(CurrentMessage) + '.MSG') then Exit;
  SetMapDot(CurrentMessage, False);
  KillMessage:=True;
 end;

{* Private methods *}

procedure TFidoMessageBase.RebuildIndex;
 var
  Find: PMessageBaseFind;
{$IFDEF VIRTUALPASCAL}
  Num, Code: Longint;
{$ELSE}
  Num, Code: Integer;
{$ENDIF}
 begin
  FillChar(Map, SizeOf(Map), 0);
  TotalMessages:=0;
  HighMessage:=0;
  Find:=CreateMessageBaseFind;
  if Find^.StartSearch(Directory + '*.MSG', faReadonly + faArchive) then
   repeat
    Val(Copy(Find^.iName, 1, Pos('.', Find^.iName) - 1), Num, Code);
    if (Code = 0) and (Num > 0) then
     begin
      SetMapDot(Num, True);

      Inc(TotalMessages);

      if Num > HighMessage then
       HighMessage:=Num;
     end;
   until not Find^.NextSearch;
  Find^.StopSearch;
  Dispose(Find, Done);
 end;

procedure TFidoMessageBase.SetMapDot(Num: Integer; Value: Boolean);
 var
  N: Integer;
 begin
  if Num >= MaxMapSize then Exit;
  N:=Num shr $03;
  if Value then
   Map[N]:=Map[N] or MapPos[Num and $07]
  else
   Map[N]:=Map[N] and not MapPos[Num and $07];
 end;

function TFidoMessageBase.GetMapDot(Num: Integer): Boolean;
 begin
  if Num >= MaxMapSize then
   begin
    GetMapDot:=False;
    Exit;
   end;
  GetMapDot:=Map[Num shr $03] and MapPos[Num and $07] <> 0;
 end;

function TFidoMessageBase.MapAttribute(var Attribute: Longint): Boolean;
 begin
  MapAttribute:=True;
  case Attribute of
   maPrivate: Attribute:=faPrivate;
   maCrash: Attribute:=faCrash;
   maReceived: Attribute:=faReceived;
   maSent: Attribute:=faSent;
   maAttach: Attribute:=faAttach;
   maTransit: Attribute:=faTransit;
   maOrphan: Attribute:=faOrphan;
   maKill: Attribute:=faKill;
   maLocal: Attribute:=faLocal;
   maHold: Attribute:=faHold;
   maFRq: Attribute:=faFRq;
   maRRq: Attribute:=faRRq;
   maRRc: Attribute:=faRRc;
   maARq: Attribute:=faARq;
   maURq: Attribute:=faURq;
  else
   MapAttribute:=False;
  end;
 end;

end.
