{$B-}
unit skMHLsq;

interface
uses
     skMHL, skCommon;

const
 smbAbsentSQD           = $3000;
 smbAbsentSQI           = $3001;
 smbCantOpenSQD         = $3002;
 smbCantOpenSQI         = $3003;
 smbIndexProblems       = $3004;
 smbUnsupported         = $3005;
 smbMessageIsNotExists  = $3006;
 smbGrungedFrame        = $3007;
 smbDataProblems        = $3008;
 smbCantCreateSQD       = $3009;
 smbCantCreateSQI       = $300A;

 squishFrameID          = $AFAE4453;
 squishFrameMessage     = $0000;
 squishFrameFree        = $0001;

 squishaPrivate         = $00001;
 squishaCrash           = $00002;
 squishaReceived        = $00004;
 squishaSent            = $00008;
 squishaAttach          = $00010;
 squishaTransit         = $00020;
 squishaOrphan          = $00040;
 squishaKill            = $00080;
 squishaLocal           = $00100;
 squishaHold            = $00200;
 squishaXX2             = $00400;
 squishaFRq             = $00800;
 squishaRRq             = $01000;
 squishaRRc             = $02000;
 squishaARq             = $04000;
 squishaURq             = $08000;
 squishaScanned         = $10000;
 squishaUID             = $20000;

 SquishMaxMsg           : Longint = $FFFF;
 SquishKeepDays         : Longint = $7FFF;

type
 PControlInformationBuffer = ^TControlInformationBuffer;
 TControlInformationBuffer = Array[1..65521] Of Char;

 TSquishBaseHeader = record
  Len: Word;
  Rsvd1: Word;
  NumMsg: LongInt;
  HighMsg: LongInt;
  SkipMsg: LongInt;
  HighWater: LongInt;
  UID: LongInt;
  Base: Array[1..80] Of Char;
  FirstFrame: LongInt;
  LastFrame: LongInt;
  FirstFree: LongInt;
  LastFree: LongInt;
  EndFrame: LongInt;
  MaxMsg: LongInt;
  KeepDays: Word;
  SqHdrSize: Word;
  Rsvd2: Array[1..124] Of Byte;
 end;

 TSquishIndex = record
  Offset: Longint;
  Number: Longint;
  Hash: Longint;
 end;

 TSquishFrame = record
  Id: LongInt;
  NextFrame: LongInt;
  PrevFrame: LongInt;
  FrameLength: LongInt;
  MsgLength: LongInt;
  ControlLength: LongInt;
  FrameType: Word;
  Rsvd: Word;
 end;

 TSquishAddress = record
  Zone, Net, Node, Point: Word;
 end;

 TSquishMessageHeader = record
  Attr: LongInt;
  MsgFrom: Array[1..36] Of Char;
  MsgTo: Array[1..36] Of Char;
  Subj: Array[1..72] Of Char;
  Orig: TSquishAddress;
  Dest: TSquishAddress;
  DateWritten: LongInt;
  DateArrived: LongInt;
  UTCoffset: Word;
  ReplyTo: LongInt;
  Replies: Array[1..9] Of LongInt;
  UID: Longint;
  AzDate: Array[1..20] Of Char;
 end;

 PSquishMessageBase = ^TSquishMessageBase;
 TSquishMessageBase = object(TMessageBase)
  constructor Init;
  destructor Done; virtual;
  function Open(const Path: String): Boolean; virtual;
  function Create(const Path: String): Boolean; virtual;
  procedure Close; virtual;
  function Exists(Message: Longint): Boolean; virtual;
  function Current: Longint; virtual;
  function GetLocation: Longint; virtual;
  procedure SetLocation(Location: Longint); virtual;
  function OpenMessage: Boolean; virtual;
  function CloseMessage: Boolean; virtual;
  function GetHighest: Longint; virtual;
  function GetCount: Longint; virtual;
  function GetFrom: String; virtual;
  function GetTo: String; virtual;
  function GetSubject: String; virtual;
  procedure SetFrom(const S: String); virtual;
  procedure SetTo(const S: String); virtual;
  procedure SetSubject(const S: String); virtual;
  function GetAttribute(Attribute: Longint): Boolean; virtual;
  procedure SetAttribute(Attribute: Longint; Enable: Boolean); virtual;
  procedure GetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure GetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  function WriteMessage: Boolean; virtual;
  function CreateNewMessage: Boolean; virtual;
  function KillMessage: Boolean; virtual;
 private
  BasePath: String;
  SquishBaseHeader: TSquishBaseHeader;
  DataLink: PMessageBaseStream;
  IndexLink: PMessageBaseStream;
  SquishIndex: TSquishIndex;
  SquishIndexNumber: Longint;
  SquishFrame: TSquishFrame;
  SquishFramePosition: Longint;
  SquishMessageHeader: TSquishMessageHeader;
  Created: Boolean;
  procedure ResetAll;
  function GetIndex(const Number: Longint; var Index: TSquishIndex): Boolean;
  function SetIndex(const Number: Longint; var Index: TSquishIndex): Boolean;
  function GetIndexSize: Longint;
  procedure SetIndexSize(const NewSize: Longint);
  function SearchIndex(const Message: Longint): Longint;
  function MapAttribute(var Attribute: Longint): Boolean;
  procedure SquishSwap(var A: Longint);
  function MakeHash(const Name: String): Longint;
  function GetFrame(const Position: Longint; var Frame: TSquishFrame): Boolean;
  function SetFrame(const Position: Longint; var Frame: TSquishFrame): Boolean;
  procedure SaveBaseHeader;
  procedure LinkToPrev(var Frame: TSquishFrame; const VictimPosition, FramePosition: Longint);
  procedure LinkToNext(var Frame: TSquishFrame; const VictimPosition, FramePosition: Longint);
  procedure LinkUpdate(var Frame: TSquishFrame; var FirstFrame, LastFrame: Longint; const FramePosition: Longint);
  procedure UnlinkFrame(var Frame: TSquishFrame; var FirstFrame, LastFrame: Longint);
 end;

implementation

constructor TSquishMessageBase.Init;
 begin
  inherited Init;
  BasePath:='';
  MessageText:=Nil;
  DataLink:=Nil;
  IndexLink:=Nil;
  CurrentMessage:=0;
 end;

destructor TSquishMessageBase.Done;
 begin
  Close;
  inherited Done;
 end;

function TSquishMessageBase.Open(const Path: String): Boolean;
 var
  K: Longint;
  I: TSquishIndex;
 begin
  if Opened then Close;
  Open:=False;
  BasePath:=Path;
  if BasePath[Length(BasePath)] <> '.' then BasePath:=Concat(BasePath, '.');
  if not FileExists(BasePath + 'SQD') then
   begin
    Status:=smbAbsentSQD;
    Exit;
   end;
  if not FileExists(BasePath + 'SQI') then
   begin
    Status:=smbAbsentSQI;
    Exit;
   end;
  DataLink:=Nil;
  IndexLink:=Nil;
  repeat
   DataLink:=CreateMessageBaseFileStream(BasePath + 'SQD', smOpen);
   if DataLink^.Status <> smOk then
    begin
     Status:=smbCantOpenSQD;
     Break;
    end;
   IndexLink:=CreateMessageBaseFileStream(BasePath + 'SQI', smOpen);
   if IndexLink^.Status <> smOk then
    begin
     Status:=smbCantOpenSQI;
     Break;
    end;
   ResetAll;
   DataLink^.Read(SquishBaseHeader, SizeOf(SquishBaseHeader));

   Open:=True;
   Opened:=True;
   Exit;
  until False;
  if DataLink <> Nil then Dispose(DataLink, Done);
  if IndexLink <> Nil then Dispose(IndexLink, Done);
  DataLink:=Nil;
  IndexLink:=Nil;
  BasePath:='';
 end;

function TSquishMessageBase.Create(const Path: String): Boolean;
 begin
  if Opened then Close;
  Create:=False;
  Opened:=False;
  BasePath:=Path;
  if BasePath[Length(BasePath)] <> '.' then BasePath:=Concat(BasePath, '.');
  repeat
   DataLink:=Nil;
   IndexLink:=Nil;
   DataLink:=CreateMessageBaseFileStream(BasePath + 'SQD', smCreate);
   if DataLink^.Status <> smOk then
    begin
     Status:=smbCantCreateSQD;
     Break;
    end;
   IndexLink:=CreateMessageBaseFileStream(BasePath + 'SQI', smCreate);
   if IndexLink^.Status <> smOk then
    begin
     Status:=smbCantCreateSQI;
     Break;
    end;
   ResetAll;
   with SquishBaseHeader do
    begin
     Len:=SizeOf(SquishBaseHeader);
     Rsvd1:=0;
     NumMsg:=0;
     HighMsg:=0;
     SkipMsg:=0;
     HighWater:=0;
     UID:=1;
     FillChar(Base, SizeOf(Base), 0);
     ToASCIIZ('$SKMHL$', @Base);
     FirstFrame:=0;
     LastFrame:=0;
     FirstFree:=0;
     LastFree:=0;
     EndFrame:=SizeOf(SquishBaseHeader);
     MaxMsg:=SquishMaxMsg;
     KeepDays:=SquishKeepDays;
     SqHdrSize:=SizeOf(SquishMessageHeader);
     FillChar(Rsvd2, SizeOf(Rsvd2), 0);
    end;
   DataLink^.Write(SquishBaseHeader, SizeOf(SquishBaseHeader));
   Create:=True;
   Opened:=True;
   Exit;
  until True;
  if DataLink <> Nil then Dispose(DataLink, Done);
  if IndexLink <> Nil then Dispose(IndexLink, Done);
  DataLink:=Nil;
  IndexLink:=Nil;
  BasePath:='';
 end;

procedure TSquishMessageBase.Close;
 begin
  if not Opened then Exit;
  Opened:=False;

  SaveBaseHeader;

  Dispose(DataLink, Done);
  Dispose(IndexLink, Done);
  DataLink:=Nil;
  IndexLink:=Nil;
 end;

function TSquishMessageBase.Exists(Message: Longint): Boolean;
 begin
  Exists:=(Message >= 1) and (Message <= SquishBaseHeader.NumMsg);
 end;

function TSquishMessageBase.Current: Longint;
 begin
  Current:=CurrentMessage;
 end;

function TSquishMessageBase.GetLocation: Longint;
 var
  I: TSquishIndex;
 begin
  GetIndex(CurrentMessage - 1, I);
  GetLocation:=I.Number;
 end;

procedure TSquishMessageBase.SetLocation(Location: Longint);
 var
  I: TSquishIndex;
  K: Longint;
 begin
  for K:=0 to GetIndexSize - 1 do
   begin
    GetIndex(K, I);
    if I.Number = Location then
     begin
      CurrentMessage:=K + 1;
      Exit;
     end;
   end;
  CurrentMessage:=0;
 end;

function TSquishMessageBase.OpenMessage: Boolean;
 var
  CIB: PControlInformationBuffer;
  CIBLength, K: Longint;
  Line: PChar;
  LineCh: Array[1..2] Of Char;
 begin
  if MessageText <> Nil then CloseMessage;

  OpenMessage:=False;
  SquishIndexNumber:=CurrentMessage - 1;
  if SquishIndexNumber = -1 then
   begin
    Status:=smbMessageIsNotExists;
    Exit;
   end;
  if not GetIndex(SquishIndexNumber, SquishIndex) then Exit;
  if not GetFrame(SquishIndex.Offset, SquishFrame) then Exit;

  SquishFramePosition:=SquishIndex.Offset;

  if SquishFrame.Id <> squishFrameID then
   begin
    Status:=smbGrungedFrame;
    Exit;
   end;

  if SquishFrame.FrameType <> squishFrameMessage then Exit;

  DataLink^.Read(SquishMessageHeader, SizeOf(SquishMessageHeader));

  CIBLength:=SquishFrame.ControlLength + 1;
  if CIBLength > SizeOf(TControlInformationBuffer) - 1 then Exit;

  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);

  GetMem(Line, MaxLineSize);
  GetMem(CIB, CIBLength);

  LineCh[2]:=#0;

  DataLink^.Read(CIB^, CIBLength - 1);
  CIB^[CIBLength]:=#1;

  ToASCIIZ('', Line);

  for K:=1 to CIBLength do
   begin
    LineCh[1]:=CIB^[K];
    if (LineCh[1] = #1) and (K <> 1) then
     begin
      LineCh[1]:=#13;
      MessageText^.Write(Line^, LenASCIIZ(Line));
      MessageText^.Write(LineCh[1], 1);
      LineCh[1]:=#1;
      ToASCIIZ('', Line);
     end;
    ConcatASCIIZ(Line, @LineCh);
   end;

  FreeMem(CIB, CIBLength);
  FreeMem(Line, MaxLineSize);

  if SquishFrame.ControlLength > SizeOf(TControlInformationBuffer) - 1 then Exit;

  MessageText^.CopyFrom(DataLink^, SquishFrame.MsgLength - SizeOf(SquishMessageHeader) - SquishFrame.ControlLength - 1);

  MessageText^.Seek(0);

  ResetAll;

  Created:=False;
  OpenMessage:=True;
 end;

function TSquishMessageBase.CloseMessage: Boolean;
 begin
  CloseMessage:=False;
  if MessageText = Nil then Exit;
  CloseMessage:=True;
  Dispose(MessageText, Done);
  MessageText:=Nil;
 end;

function TSquishMessageBase.GetHighest: Longint;
 begin
  GetHighest:=SquishBaseHeader.NumMsg;
 end;

function TSquishMessageBase.GetCount: Longint;
 begin
  GetCount:=SquishBaseHeader.NumMsg;
 end;

function TSquishMessageBase.GetFrom: String;
 begin
  GetFrom:=FromASCIIZ(@SquishMessageHeader.MsgFrom);
 end;

function TSquishMessageBase.GetTo: String;
 begin
  GetTo:=FromASCIIZ(@SquishMessageHeader.MsgTo);
 end;

function TSquishMessageBase.GetSubject: String;
 begin
  GetSubject:=FromASCIIZ(@SquishMessageHeader.Subj);
 end;

procedure TSquishMessageBase.SetFrom(const S: String);
 begin
  ToASCIIZ(S, @SquishMessageHeader.MsgFrom);
 end;

procedure TSquishMessageBase.SetTo(const S: String);
 begin
  ToASCIIZ(S, @SquishMessageHeader.MsgTo);
 end;

procedure TSquishMessageBase.SetSubject(const S: String);
 begin
  ToASCIIZ(S, @SquishMessageHeader.Subj);
 end;

function TSquishMessageBase.GetAttribute(Attribute: Longint): Boolean;
 begin
  if not MapAttribute(Attribute) then
   begin
    GetAttribute:=False;
    Exit;
   end;
  GetAttribute:=SquishMessageHeader.Attr and Attribute <> 0;
 end;

procedure TSquishMessageBase.SetAttribute(Attribute: Longint; Enable: Boolean);
 begin
  if not MapAttribute(Attribute) then Exit;
  if Enable then
   SquishMessageHeader.Attr:=SquishMessageHeader.Attr or Attribute
  else
   SquishMessageHeader.Attr:=SquishMessageHeader.Attr and not Attribute
 end;

procedure TSquishMessageBase.GetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 var
  A: Longint;
 begin
  A:=SquishMessageHeader.DateWritten;
  SquishSwap(A);
  DosDateTimeToMessageBaseDateTime(A, DateTime);
 end;

procedure TSquishMessageBase.GetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 var
  A: Longint;
 begin
  A:=SquishMessageHeader.DateArrived;
  SquishSwap(A);
  DosDateTimeToMessageBaseDateTime(A, DateTime);
 end;

procedure TSquishMessageBase.SetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 begin
  MessageBaseDateTimeToDosDateTime(DateTime, SquishMessageHeader.DateWritten);
  SquishSwap(SquishMessageHeader.DateWritten);
 end;

procedure TSquishMessageBase.SetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
  MessageBaseDateTimeToDosDateTime(DateTime, SquishMessageHeader.DateArrived);
  SquishSwap(SquishMessageHeader.DateArrived);
 end;

function TSquishMessageBase.WriteMessage: Boolean;
 function lz(const Number: Longint): String;
  var
   S: String[2];
  begin
   Str(Number, S);

   if S[0] = #1 then
    S:=Concat('0', S);

   lz:=S;
  end;
 var
  CIB: PControlInformationBuffer;
  CIBLength, TextStart: Longint;
  Line: PChar;
  FromAddress, ToAddress: TAddress;
  DateTime: TMessageBaseDateTime;
  TempSquishFrame: TSquishFrame;
 const
  C: Char = #0;
 begin
  WriteMessage:=False;
  GetMem(Line, MaxLineSize);
  New(CIB);

  if CurrentMessage > SquishBaseHeader.NumMsg then
   begin
    SquishBaseHeader.NumMsg:=CurrentMessage;
    SquishBaseHeader.HighMsg:=CurrentMessage;
   end;

  MessageText^.Seek(0);

  CIBLength:=1;
  TextStart:=0;

  while not EndOfMessage do
   begin
    TextStart:=MessageText^.GetPos;

    GetStringPChar(Line, MaxLineSize);

    if Line[0] <> #1 then
     Break;

    Move(Line[0], CIB^[CIBLength], LenASCIIZ(Line));

    Inc(CIBLength, LenASCIIZ(Line));
   end;

  if (TextStart = 0) and (Line[0] = #1) then
   TextStart:=MessageText^.GetPos;

  CIB^[CIBLength]:=#0;

  TempSquishFrame:=SquishFrame;

  SquishFrame.MsgLength:=SizeOf(TSquishMessageHeader) + CIBLength + (MessageText^.GetSize - TextStart) + 1;
  SquishFrame.FrameLength:=SquishFrame.MsgLength;

  if Created then
   begin
    SquishFrame.PrevFrame:=SquishBaseHeader.LastFrame;
    SquishFrame.NextFrame:=0;
   end;

  if TempSquishFrame.FrameLength >= SquishFrame.FrameLength then
   SquishFrame.FrameLength:=TempSquishFrame.FrameLength
  else
   begin
    if not Created then
     begin
      UnlinkFrame(TempSquishFrame, SquishBaseHeader.FirstFrame, SquishBaseHeader.LastFrame);

      TempSquishFrame.NextFrame:=0;
      TempSquishFrame.FrameType:=SquishFrameFree;

      LinkToPrev(TempSquishFrame, SquishBaseHeader.LastFree, SquishFramePosition);
      LinkUpdate(TempSquishFrame, SquishBaseHeader.FirstFree, SquishBaseHeader.LastFree, SquishFramePosition);

      SetFrame(SquishFramePosition, TempSquishFrame);
     end;

    SquishFramePosition:=SquishBaseHeader.FirstFree;

    while SquishFramePosition <> 0 do
     begin
      GetFrame(SquishFramePosition, TempSquishFrame);

      if TempSquishFrame.FrameLength >= SquishFrame.MsgLength then
       Break;

      SquishFramePosition:=TempSquishFrame.NextFrame;
     end;
   end;

  if SquishFramePosition = 0 then
   begin
    SquishFramePosition:=SquishBaseHeader.EndFrame;

    Inc(SquishBaseHeader.EndFrame, SquishFrame.FrameLength + SizeOf(TSquishFrame));
   end
  else
   begin
    SquishFrame.FrameLength:=TempSquishFrame.FrameLength;

    UnlinkFrame(TempSquishFrame, SquishBaseHeader.FirstFree, SquishBaseHeader.LastFree);
   end;

  SquishIndex.Offset:=SquishFramePosition;
  SquishIndex.Hash:=MakeHash(GetTo);

  SquishFrame.ID:=SquishFrameID;
  SquishFrame.ControlLength:=CIBLength;
  SquishFrame.FrameType:=SquishFrameMessage;
  SquishFrame.Rsvd:=0;

  LinkToNext(SquishFrame, SquishFrame.NextFrame, SquishFramePosition);
  LinkToPrev(SquishFrame, SquishFrame.PrevFrame, SquishFramePosition);
  LinkUpdate(SquishFrame, SquishBaseHeader.FirstFrame, SquishBaseHeader.LastFrame, SquishFramePosition);

  GetFromAndToAddress(FromAddress, ToAddress);

  SquishMessageHeader.Orig.Zone:=FromAddress.Zone;
  SquishMessageHeader.Orig.Net:=FromAddress.Net;
  SquishMessageHeader.Orig.Node:=FromAddress.Node;
  SquishMessageHeader.Orig.Point:=FromAddress.Point;
  SquishMessageHeader.Dest.Zone:=ToAddress.Zone;
  SquishMessageHeader.Dest.Net:=ToAddress.Net;
  SquishMessageHeader.Dest.Node:=ToAddress.Node;
  SquishMessageHeader.Dest.Point:=ToAddress.Point;
  SquishMessageHeader.UTCOffset:=0;

  GetWrittenDateTime(DateTime);

  ToASCIIZ(lz(DateTime.Day) + ' ' + MonthNumberToMonthString(DateTime.Month) + ' ' + lz(DateTime.Year - 1900) + '  ' +
   lz(DateTime.Hour) + ':' + lz(DateTime.Min) + ':' + lz(DateTime.Sec), @SquishMessageHeader.AzDate);

  SquishIndexNumber:=CurrentMessage - 1;

  SetIndex(SquishIndexNumber, SquishIndex);

  MessageText^.Seek(TextStart);

  DataLink^.Seek(SquishFramePosition);
  DataLink^.Write(SquishFrame, SizeOf(SquishFrame));
  DataLink^.Write(SquishMessageHeader, SizeOf(SquishMessageHeader));
  DataLink^.Write(CIB^, CIBLength);
  DataLink^.CopyFrom(MessageText^, MessageText^.GetSize - TextStart);
  DataLink^.Write(C, SizeOf(C));

  Dispose(CIB);
  FreeMem(Line, MaxLineSize);

  SaveBaseHeader;

  WriteMessage:=True;
 end;

function TSquishMessageBase.CreateNewMessage: Boolean;
 begin
  FillChar(SquishMessageHeader, SizeOf(SquishMessageHeader), 0);

  CurrentMessage:=SquishBaseHeader.NumMsg + 1;
  SquishMessageHeader.UID:=SquishBaseHeader.UID;
  Inc(SquishBaseHeader.UID);
  SquishFramePosition:=0;
  SquishMessageHeader.Attr:=squishaUID;

  FillChar(SquishFrame, SizeOf(SquishFrame), 0);
  SquishFrame.ID:=SquishFrameID;

  FillChar(SquishIndex, SizeOf(SquishIndex), 0);
  SquishIndex.Number:=SquishMessageHeader.UID;

  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);

  Created:=True;
  CreateNewMessage:=True;
 end;

function TSquishMessageBase.KillMessage: Boolean;
 var
  K: Longint;
  I: TSquishIndex;
 begin
  UnlinkFrame(SquishFrame, SquishBaseHeader.FirstFrame, SquishBaseHeader.LastFrame);

  for K:=CurrentMessage to GetIndexSize - 1 do
   begin
    GetIndex(K, I);
    SetIndex(K - 1, I);
   end;
  SetIndexSize(GetIndexSize - 1);
  Dec(CurrentMessage);

  SquishFrame.NextFrame:=0;
  SquishFrame.FrameType:=SquishFrameFree;
  LinkToPrev(SquishFrame, SquishBaseHeader.LastFree, SquishFramePosition);
  LinkUpdate(SquishFrame, SquishBaseHeader.FirstFree, SquishBaseHeader.LastFree, SquishFramePosition);
  SetFrame(SquishFramePosition, SquishFrame);
  Dec(SquishBaseHeader.NumMsg);
  Dec(SquishBaseHeader.HighMsg);
  SaveBaseHeader;
  KillMessage:=True;
 end;

{* Private methods *}

procedure TSquishMessageBase.ResetAll;
 begin
  DataLink^.Reset;
  DataLink^.Flush;
  IndexLink^.Reset;
  IndexLink^.Flush;
 end;

function TSquishMessageBase.GetIndex(const Number: Longint; var Index: TSquishIndex): Boolean;
 begin
  Status:=smbIndexProblems;
  GetIndex:=False;
  IndexLink^.Reset;
  IndexLink^.Seek(Number * SizeOf(TSquishIndex));
  if IndexLink^.Status <> smOk then Exit;
  IndexLink^.Read(Index, SizeOf(Index));
  if IndexLink^.Status <> smOk then Exit;
  GetIndex:=True;
 end;

procedure TSquishMessageBase.SetIndexSize(const NewSize: Longint);
 begin
  IndexLink^.Reset;
  IndexLink^.Seek(NewSize * SizeOf(TSquishIndex));
  IndexLink^.Truncate;
 end;

function TSquishMessageBase.SetIndex(const Number: Longint; var Index: TSquishIndex): Boolean;
 begin
  Status:=smbIndexProblems;
  SetIndex:=False;
  IndexLink^.Reset;
  IndexLink^.Seek(Number * SizeOf(TSquishIndex));
  if IndexLink^.Status <> smOk then Exit;
  IndexLink^.Write(Index, SizeOf(Index));
  if IndexLink^.Status <> smOk then Exit;
  SetIndex:=True;
 end;

function TSquishMessageBase.GetIndexSize: Longint;
 begin
  IndexLink^.Reset;
  GetIndexSize:=IndexLink^.GetSize div SizeOf(TSquishIndex);
 end;

function TSquishMessageBase.SearchIndex(const Message: Longint): Longint;
 var
  K: Longint;
  I: TSquishIndex;
 begin
  for K:=0 to GetIndexSize - 1 do
   begin
    GetIndex(K, I);
    if I.Number = Message then
     begin
      SearchIndex:=K;
      Exit;
     end;
   end;
  SearchIndex:=-1;
 end;

function TSquishMessageBase.MapAttribute(var Attribute: Longint): Boolean;
 begin
  MapAttribute:=True;
  case Attribute of
   maPrivate: Attribute:=squishaPrivate;
   maCrash: Attribute:=squishaCrash;
   maReceived: Attribute:=squishaReceived;
   maSent: Attribute:=squishaSent;
   maAttach: Attribute:=squishaAttach;
   maTransit: Attribute:=squishaTransit;
   maOrphan: Attribute:=squishaOrphan;
   maKill: Attribute:=squishaKill;
   maLocal: Attribute:=squishaLocal;
   maHold: Attribute:=squishaHold;
   maFRq: Attribute:=squishaFRq;
   maRRq: Attribute:=squishaRRq;
   maARq: Attribute:=squishaARq;
  else
   MapAttribute:=False;
  end;
 end;

procedure TSquishMessageBase.SquishSwap(var A: Longint);
 begin
  A:=(A shr 16) + ((A and $FFFF) shl 16);
 end;

function TSquishMessageBase.MakeHash(const Name: String): Longint;
 var
  K: Byte;
  Hash, Tmp: Longint;
 begin
  Hash:=0;
  for K:=1 to Length(Name) do
   begin
    Hash:=(Hash shl 4) + Byte(LoCase(Name[K]));
    Tmp:=Hash and $F0000000;
    if Tmp <> 0 then
     Hash:=(Hash or (Tmp shr 24)) or Tmp;
   end;
  MakeHash:=Hash and $7FFFFFFF;
 end;

function TSquishMessageBase.GetFrame(const Position: Longint; var Frame: TSquishFrame): Boolean;
 begin
  Status:=smbDataProblems;
  GetFrame:=False;
  if Position = 0 then Exit;
  DataLink^.Reset;
  DataLink^.Seek(Position);
  if DataLink^.Status <> smOk then Exit;
  DataLink^.Read(Frame, SizeOf(Frame));
  if DataLink^.Status <> smOk then Exit;
  GetFrame:=True;
 end;

function TSquishMessageBase.SetFrame(const Position: Longint; var Frame: TSquishFrame): Boolean;
 begin
  Status:=smbDataProblems;
  SetFrame:=False;
  DataLink^.Reset;
  DataLink^.Seek(Position);
  if DataLink^.Status <> smOk then Exit;
  DataLink^.Write(Frame, SizeOf(Frame));
  if DataLink^.Status <> smOk then Exit;
  SetFrame:=True;
 end;

procedure TSquishMessageBase.SaveBaseHeader;
 begin
  DataLink^.Seek(0);
  DataLink^.Write(SquishBaseHeader, SizeOf(SquishBaseHeader));
 end;

procedure TSquishMessageBase.LinkToPrev(var Frame: TSquishFrame; const VictimPosition, FramePosition: Longint);
 var
  TheVictim: TSquishFrame;
 begin
  if GetFrame(VictimPosition, TheVictim) then
   begin
    TheVictim.NextFrame:=FramePosition;
    Frame.PrevFrame:=VictimPosition;
    SetFrame(VictimPosition, TheVictim);
   end
  else
   Frame.PrevFrame:=0;
 end;

procedure TSquishMessageBase.LinkToNext(var Frame: TSquishFrame; const VictimPosition, FramePosition: Longint);
 var
  TheVictim: TSquishFrame;
 begin
  if GetFrame(VictimPosition, TheVictim) then
   begin
    TheVictim.PrevFrame:=FramePosition;
    Frame.NextFrame:=VictimPosition;
    SetFrame(VictimPosition, TheVictim);
   end
  else
   Frame.NextFrame:=0;
 end;

procedure TSquishMessageBase.LinkUpdate(var Frame: TSquishFrame; var FirstFrame, LastFrame: Longint;
  const FramePosition: Longint);
 begin
  if Frame.NextFrame = 0 then LastFrame:=FramePosition;
  if Frame.PrevFrame = 0 then FirstFrame:=FramePosition;
 end;

procedure TSquishMessageBase.UnlinkFrame(var Frame: TSquishFrame; var FirstFrame, LastFrame: Longint);
 var
  TempSquishFrame: TSquishFrame;
 begin
  if GetFrame(Frame.PrevFrame, TempSquishFrame) then
   begin
    LinkToNext(TempSquishFrame, Frame.NextFrame, Frame.PrevFrame);
    SetFrame(Frame.PrevFrame, TempSquishFrame);
   end
  else
   FirstFrame:=Frame.NextFrame;
  if GetFrame(Frame.NextFrame, TempSquishFrame) then
   begin
    LinkToPrev(TempSquishFrame, Frame.PrevFrame, Frame.NextFrame);
    SetFrame(Frame.NextFrame, TempsquishFrame);
   end
  else
   LastFrame:=Frame.PrevFrame;
 end;

end.