unit skOpen;

interface
uses
     skMHL,
     skMHLmsg, skMHLjam, skMHLsq,
     skCommon;

const
 ombUnknownIDorUnsupported      = $A000;

var
 OpenStatus: Longint;

function OpenMessageBase(var Base: PMessageBase; const ID: String): Boolean;
function OpenOrCreateMessageBase(var Base: PMessageBase; const ID: String): Boolean;
function CloseMessageBase(var Base: PMessageBase): Boolean;

function InitMessageBase(var Base: PMessageBase; const ID: String): Boolean;
function DoneMessageBase(var Base: PMessageBase): Boolean;

implementation

type
 TMessageBaseFormat = (mbfJam, mbfMSG, mbfSquish, mbfUnknown);

procedure SplitID(const ID: String; var Format: TMessageBaseFormat; var Path: String);
 begin
  if Length(ID) = 0 then
   begin
    Format:=mbfUnknown;
    Exit;
   end;
  case ID[1] of
   'f', 'F', 'm', 'M', '*': Format:=mbfMSG;
   'j', 'J': Format:=mbfJam;
   's', 'S': Format:=mbfSquish;
  else
   Format:=mbfUnknown;
  end;
  case Format of
   mbfMSG, mbfJam, mbfSquish:
    Path:=Copy(ID, 2, 255);
  else
   Path:=ID;
  end;
 end;

function OpenMessageBase(var Base: PMessageBase; const ID: String): Boolean;
 var
  Format: TMessageBaseFormat;
  Path: String;
 begin
  if InitMessageBase(Base, ID) then
   begin
    SplitID(ID, Format, Path);
    if not Base^.Open(Path) then
     begin
      OpenStatus:=Base^.Status;
      OpenMessageBase:=False;
      DoneMessageBase(Base);
     end
    else
     OpenMessageBase:=True;
   end
  else
   OpenMessageBase:=False;
 end;

function OpenOrCreateMessageBase(var Base: PMessageBase; const ID: String): Boolean;
 var
  Format: TMessageBaseFormat;
  Path: String;
 begin
  if InitMessageBase(Base, ID) then
   begin
    SplitID(ID, Format, Path);
    if not Base^.Open(Path) then
     if not Base^.Create(Path) then
      begin
       OpenStatus:=Base^.Status;
       OpenOrCreateMessageBase:=False;
       Exit;
      end;
    OpenOrCreateMessageBase:=True;
   end
  else
   OpenOrCreateMessageBase:=False;
 end;

function CloseMessageBase(var Base: PMessageBase): Boolean;
 begin
  if Base <> Nil then
   begin
    Base^.Close;
    CloseMessageBase:=DoneMessageBase(Base);
   end
  else
   CloseMessageBase:=False;
 end;

function InitMessageBase(var Base: PMessageBase; const ID: String): Boolean;
 var
  Format: TMessageBaseFormat;
  Path: String;
 begin
  Base:=Nil;
  InitMessageBase:=False;
  OpenStatus:=ombUnknownIDorUnsupported;
  if Length(ID) = 0 then Exit;
  SplitID(ID, Format, Path);
  case Format of
   mbfMSG: Base:=New(PFidoMessageBase, Init);
   mbfSquish: Base:=New(PSquishMessageBase, Init);
   mbfJam: Base:=New(PJamMessageBase, Init);
  else
   Exit;
  end;
  InitMessageBase:=True;
 end;

function DoneMessageBase(var Base: PMessageBase): Boolean;
 begin
  DoneMessageBase:=Base <> Nil;
  if Base <> Nil then
   Dispose(Base, Done);
  Base:=Nil;
 end;

end.